

/*{{{  Header*/
/*------------------------------------------------------------------*/
/*                                                                  */
/*                            socket.c                              */
/*                                                                  */
/*     by Nick Clarke, Southampton University Computing Services    */
/*                                                                  */
/*     Functions added to iserver in order to implement a BSD 4.3   */
/*     socket library on the root transputer.                       */
/*                                                                  */
/*     Tag choices correspond to Inmos B300 (TCPLINK) tags.         */
/*                                                                  */
/*------------------------------------------------------------------*/
/*}}}*/
/*{{{  Includes, Constants and Macros*/
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <process.h>
#include <pwd.h>
void free(void *);
#pragma IMS_translate(server_transaction,"server_transaction%c")
extern int server_transaction(char *, int, char *);

static enum {
   SpBuffSize=512,
   EnvSize=500,
   NoEnv=4
   };

#define PACK_ITEM(x,y,c) { char *n=(char *)&(x); int i; \
        for(i=0;i<sizeof(x);i++) (y)[c++]=*(n++); }
#define PACK_SLICE(x,l,y,c) (y)[c++]=(l)&255; (y)[c++]=((l)>>8)&255; \
                            _memcpy((y)+c,(x),l); c+=l;
#define UNPACK_ITEM(x,y,c) { char *n=(char *)&(x); int i; \
        for(i=0;i<sizeof(x);i++) *(n++)=(y)[c++]; }
#define UNPACK_SLICE(x,l,y,c) (l)=(y)[c++]; (l)+=((y)[c++])<<8; \
                              _memcpy((x),(y)+c,l); c+=l;

/*}}}*/
/*{{{  Structures defs and Statics*/
/* Used for temporary store of env stuff */
static char PseudoEnv[NoEnv][EnvSize];
static int CurrEnv=0;
/*}}}*/
/*{{{  Iserver socket tags (same as used by B300)*/
static enum Tags {
    };
/*}}}*/
/*{{{  global vars*/
char *sys_errlist[]={"Transputer X library : error messages absent",
                      NULL };
int   sys_nerr=1;
static char Send[SpBuffSize],Recv[SpBuffSize];
/*}}}*/
/*{{{  open()*/
int open(char *path, int flags, int mode)
{
  int rc, c, Len, OutSigio, SockId, NewErrno;

  /*{{{  Fill send buffer*/
  c=0;
  Send[c++]=SpSocktag1;
  Send[c++]=SpSockopen;
  PACK_SLICE(path,strlen(path),Send,c)
  PACK_ITEM(flags,Send,c)
  PACK_ITEM(mode,Send,c)
  c=(c+1) & (~1);
  /*}}}*/
  if ((Len=server_transaction(Send,c,Recv))<0)
     { errno=EIO; return(-1); }
  /*{{{  Empty receive buffer*/
  c=0;
  rc=Recv[c++];
  UNPACK_ITEM(NewErrno,Recv,c)
  UNPACK_ITEM(OutSigio,Recv,c)
  if (OutSigio) raise((OutSigio==1) ? SIGIO : SIGURG);
  UNPACK_ITEM(SockId,Recv,c)
  /*}}}*/
  if (rc && NewErrno) errno=NewErrno;
  rc=(rc==0) ? SockId : (-1);
  return(rc);
}
/*}}}*/
/*{{{  creat()*/
