/*{{{  includes and defines*/
#include <stdio.h>
#include "inmos.h"
#include "isocket.h"
#include "vcpdefs.h"
#include "systype.h"
#include "queue.h"
#include "vcptype.h"
#include "iserver.h"

#define BOOT_STRAP1	"primary"
#define BOOT_STRAP2	"secondary"
#define T9BOOT_STRAP	"t9boot"
#define WORKSPACE_SIZE	1024
#define BOOTSTRAP_LENGTH	255
/*}}}*/

/*{{{  BootStrapT8*/
/*
 *------------------------------------------------------------------------
 *
 *	BootStrapT8 :
 *		Copy the boot file to the link: used to load primary 
 *		and secondary bootstrap codes. The file size is sent first,
 *		followed by the file.
 *	Parameters :
 *		TheLink : Link number over which file is booted
 *		BootFileName : Name of file to be sent
 *		LenSize : Number of bytes in which the file size is sent
 *		ServerInfo : structure containing information about server
 *
 *	Return value :
 *		Number of bytes that could be successfully copied.
 *
 *------------------------------------------------------------------------
 */
int     BootStrapT8(TheLink, BootFileName, LenSize, ServerInfo)
int TheLink;
char *BootFileName;
int LenSize;
SERVER_INFO *ServerInfo;
{
	FILE            *fd;
	char            buffer[BOOT_BUFFER_LENGTH];
	int             length = 0;
	long            size = 0;
	static char     Flicks[] = "|/-\\|/-\\";
	int             flick = 0;
	unsigned char   ops_res;
	char		   lenbuf[1];

	/* Open the boot file */
	DEBUG(("Booting file %s", BootFileName ));
	fd = fopen(BootFileName, "rb");
	if (fd == NULL) {
		ABORT(MISC_EXIT, (SE, "Boot: cannot find file \"%s\"", BootFileName));
	}
	fseek( fd, 0, 2 ); 
	length = ftell(fd);
	fseek( fd,0,0 );

	/* Read the file in large chunks */
	while ((length = fread(buffer, 1, BOOTSTRAP_LENGTH, fd)) > 0) {
		DEBUG(("%d", length));
		if ( LenSize == 1 ){
			lenbuf[0] = (unsigned char)length;
			ops_res = WriteLink(TheLink, lenbuf, 1, BOOT_TIMEOUT );
		}
		else {
			lenbuf[0] = length;
			ops_res = WriteLink(TheLink, lenbuf, 1, BOOT_TIMEOUT );
		}

		ops_res = WriteLink(TheLink, buffer, (unsigned long) length, BOOT_TIMEOUT);

		if (ops_res <= 0) {
			ABORT(MISC_EXIT, (SE, "Boot: write failed after %d bytes \n", size));
		}

		size += length;

	}
	if ( LenSize == 2 ){
		lenbuf[0] = 0;
		ops_res = WriteLink(TheLink, lenbuf, 1, BOOT_TIMEOUT );
	}

	fclose(fd);

	DEBUG(("booted %ld bytes", size));
	return(size);
}
/*}}}*/

/*{{{  BootStrapT9*/
/*
 *------------------------------------------------------------------------
 *
 *	BootStrapT9 :
 *		Copy the boot file to a virtual channel
 *	Parameters :
 *		TheLink : Virtual link number over which file is booted
 *		BootFileName : Name of file to be sent
 *		ServerInfo : structure containing information about server
 *
 *	Return value :
 *		Number of bytes that could be successfully copied.
 *
 *------------------------------------------------------------------------
 */
int     BootStrapT9(TheLink, BootFileName, ServerInfo)
int TheLink;
char *BootFileName;
SERVER_INFO *ServerInfo;
{
	FILE            *fd;
	char            *buffer;
	int             length = 0;
	long            size = 0;
	static char     Flicks[] = "|/-\\|/-\\";
	int             flick = 0;
	unsigned char   ops_res;

	/* Open the boot file */
	DEBUG(( "Booting file %s", BootFileName ));
	fd = fopen(BootFileName, "rb");
	if (fd == NULL) {
		ABORT(MISC_EXIT, (SE, "Boot: cannot find file \"%s\"", BootFileName));
	}
	fseek( fd, 0, 2 ); 
	length = ftell(fd);
	fseek( fd,0,0 );
	buffer = (char *)malloc( length * sizeof(char));
	if ( buffer == NULL ){
		ABORT(MISC_EXIT, (SE, "Boot: cannot find file \"%s\"", BootFileName));
	}

	/* Read the file in large chunks */
	while ((length = fread(buffer, 1, length, fd)) > 0) {
		DEBUG(("%d", length));
		ops_res = WriteLink(TheLink, buffer, (unsigned long) length, -1/*BOOT_TIMEOUT*/);

		if (ops_res <= 0) {
			ABORT(MISC_EXIT, (SE, "Boot: write failed after %d bytes \n", size));
		}

		size += length;

	}

	fclose(fd);

	DEBUG(("booted %ld bytes", size));
	return(size);
}
/*}}}*/

/*{{{  CoffBootT800*/

/*
 *------------------------------------------------------------------------
 *
 *	CoffBootT800 :
 *		Load a COFF file over a link to a T8. The primary
 *		and secondary boot codes are first loaded, followed by
 *		the various sections of the COFF file.
 *	Parameters :
 *		TheLink : Link number over which file is booted
 *		ServerInfo : structure containing information about server
 *		BootFileName : Name of file to be sent
 *
 *	Returns nothing
 *
 *------------------------------------------------------------------------
 */
void CoffBootT800(int TheLink, SERVER_INFO *ServerInfo, char *BootFileName)
{
	unsigned long MemStart;
	int BootStrapLen = WORKSPACE_SIZE;
	int res;
	char *PathName;
	char *BootStrapPath;

	DEBUG(("Booting root transputer..."));
	BootStrapPath = getenv( "XPDS_SEARCH" );
	DEBUG((  "BootStrapPath %s", BootStrapPath ));
	PathName = (char *)malloc(strlen(BootStrapPath)+strlen(BOOT_STRAP1)+2);
	sprintf( PathName, "%s/%s", BootStrapPath, BOOT_STRAP1 );
	BootStrapLen += BootStrapT8( TheLink, PathName, 1, ServerInfo );
	res = ReadLink( TheLink, &MemStart, sizeof(unsigned long), -1);
	DEBUG((  "MemStart %x Address %x", MemStart, MemStart+BootStrapLen ));
	free( PathName );
	PathName = (char *)malloc(strlen(BootStrapPath)+strlen(BOOT_STRAP2)+2);
	sprintf(PathName, "%s/%s", BootStrapPath, BOOT_STRAP2 );
	BootStrapLen += BootStrapT8( TheLink, PathName, 2, ServerInfo );
	free( PathName );

	CoffLoad( BootFileName, BootStrapLen+MemStart, TheLink, ServerInfo );
	DEBUG(( "Boot COFF file over, Loader %08x Address %08x", MemStart, BootStrapLen+MemStart ));
}
/*}}}*/

/*{{{  CoffBootT9000*/
/*
 *------------------------------------------------------------------------
 *
 *	CoffBootT9000 :
 *		Load a COFF file over a virtual link to a T9. The 
 *		bootstrap code is first loaded, followed by
 *		the various sections of the COFF file. It is assumed that
 *		the T9 has been configured, and the primary bootstrap is
 *		already resident on the T9.
 *	Parameters :
 *		ServerInfo : structure containing information about server
 *		BootFileName : Name of file to be sent
 *
 *	Returns nothing
 *
 *------------------------------------------------------------------------
 */
void CoffBootT9000( SERVER_INFO *ServerInfo,char *BootFileName )
{
	int res;
	char *PathName;
	char *BootStrapPath;
	int TheLink;

	DEBUG(("Booting root transputer..."));
	TheLink = ServerInfo->ServerId;
#if 0
	BootStrapPath = getenv( "XPDS_SEARCH" );
	DEBUG((  "BootStrapPath %s TheLink %d", BootStrapPath, TheLink ));
	PathName = (char *)malloc(strlen(BootStrapPath)+strlen(T9BOOT_STRAP)+5);
	sprintf(PathName, "%s/%s", BootStrapPath, T9BOOT_STRAP );
	BootStrapT9( TheLink, PathName, ServerInfo );
	free( PathName );
#endif
	CoffLoad( BootFileName, TheLink, ServerInfo );
	DEBUG(( "Boot COFF file over" ));
}
/*}}}*/
