/*{{{  include files*/
#include <stdio.h>
#ifdef INMOSC
#include <channel.h>
#include <process.h>
#include <semaphor.h>
#else
#include <conc.h>
#endif
#include <string.h>
#include <stdlib.h>
#include "vcptype.h"
#include "tclVcpDefs.h"
#include "tclVcpType.h"
#include "queue.h"
#include "vcpmac.h"
#include "vcpext.h"
#include "tclVcpMac.h"
#include "tclVcpExt.h"
#include "debug.h"


extern Boolean VCP_Initialized;
extern Boolean VCP_Running[];

static Channel *InChannel[] = { LINK0IN, LINK1IN, LINK2IN, LINK3IN };
static Channel *OutChannel[] = { LINK0OUT, LINK1OUT, LINK2OUT, LINK3OUT };

/*}}}*/

/*{{{  Chan_BlockedReceive*/

/*
 *------------------------------------------------------------------------
 *
 *	Chan_BlockedReceive :
 *		Receive a message and store it in the specified file or
 *	in a buffer.
 *	
 *	Parameters :
 *		LinkNumber :
 *			Link on which message is to be received.
 *		FileOrBuffer :
 *			FileFlag = True implies receive into a file. In this
 *			case, BufferPtr specifies the filename.
 *			If FileFlag = False, the message is received into a
 *			buffer and the address of the buffer is returned in
 *			BufferPtr. Message is returned in a string of the
 *			form (<Hexbyte> )*
 *		count :
 *			Number of bytes to be received.
 *		TimeOut :
 *			Number of microseconds to wait for message
 *		TimeOutCommand :
 *			TCL Command to be executed on timeout
 *
 *	Returns nothing
 *
 *------------------------------------------------------------------------
 */
int Chan_BlockedReceive( int LinkNumber, FILE_OR_BUFFER *FileOrBuffer, 
	int count, int TimeOut, char *TimeOutCommand )
{
	FILE	*fp;
	char	*tempBuffer;
	char 	*MsgBuffer;

	if (( VCP_Initialized )&&( VCP_Running[LinkNumber] ))
		return(VCP_AlreadyRunning);
	tempBuffer = ( char *)vcp_malloc( count );
	if ( tempBuffer == NULL )
		return(OutOfMemory);
	if ( TimeOut > 0 ){
		if (ChanInTimeFail( InChannel[LinkNumber], tempBuffer, count, TimeOut ) != 0 )
			QUEUE_TCL_COMMAND( TimeOutCommand );
	}
	else
		ChanIn( InChannel[LinkNumber], tempBuffer, count );
	if ( FileOrBuffer->FileFlag ) {
		fp = fopen( FileOrBuffer->BufferPtr, "w" );
		DEBUG_INFO(( "Receive into file %s fp %x\n", FileOrBuffer->BufferPtr, fp ));
		fwrite( tempBuffer, count, sizeof(char), fp );
		fclose(fp);
		vcp_free( tempBuffer );
	}
	else {
		FileOrBuffer->BufferPtr = tempBuffer;
		if ( count == 0 ){
			MsgBuffer = vcp_malloc( sizeof(char));
			MsgBuffer[0] = '\0';
		}
		else
			MsgBuffer  = (char *)vcp_malloc( count * LENGTH_PER_BYTE+1 );
		FileOrBuffer->BufferPtr = MsgBuffer;
		MESSAGE_COPY( MsgBuffer, tempBuffer, count );
		vcp_free( tempBuffer );
	}
	return(VCP_OK);
}

/*}}}*/

/*{{{  Chan_ReceiveThread*/
/*
 *------------------------------------------------------------------------
 *
 *	Chan_ReceiveThread :
 *		Invokes Chan_BlockedReceive to receive a message into a file
 *	or buffer. When the receive is complete, this thread executes the
 *	TCL command specified in the parameter Command, and then terminates.
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static void Chan_ReceiveThread( int FIXME, int LinkNumber, FILE_OR_BUFFER *FileOrBuffer, int count, char *Command, char *RecvVar , int TimeOut, char *TimeOutCommand)
{
	char *SetRecvVar;

	Chan_BlockedReceive( LinkNumber, FileOrBuffer, count, TimeOut, TimeOutCommand );
	/* Set TCL variable if required */
	if ( !FileOrBuffer->FileFlag ){
		int len;
		len = strlen( RecvVar )+ strlen( FileOrBuffer->BufferPtr ) +9;
		SetRecvVar = vcp_malloc(len *sizeof(char));
		sprintf( SetRecvVar, "set %s \"%s\"", RecvVar, FileOrBuffer->BufferPtr );
		QUEUE_TCL_COMMAND( SetRecvVar );
	}
	QUEUE_TCL_COMMAND( Command );
	if ( FileOrBuffer->FileFlag )
		vcp_free( FileOrBuffer->BufferPtr );
	vcp_free( FileOrBuffer );
	vcp_free( Command ); if ( TimeOutCommand != NULL )vcp_free( TimeOutCommand ); vcp_free( RecvVar );
}
/*}}}*/

/*{{{  Chan_UnblockedReceive*/

/*
 *------------------------------------------------------------------------
 *
 *	Chan_UnBlockedReceive :
 *		Saves all the parameters and creates a thread to receive
 *	a message
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
int Chan_UnBlockedReceive( int LinkNumber, FILE_OR_BUFFER *FileOrBuffer, 
	int count, char *Command, char *RecvVar , int TimeOut, char *TimeOutCommand)
{
	Process *pid;
	char *tempCommand;
	FILE_OR_BUFFER *tempFileOrBuffer;
	char *tempFileName;
	char *tempRecvVar;
	char *tempTimeOutCommand;

	if (( VCP_Initialized )&&( VCP_Running[LinkNumber] ))
		return(VCP_AlreadyRunning);
	tempCommand = (char *)vcp_malloc( strlen(Command)+1 );
	tempFileOrBuffer = MEM_ALLOC( FILE_OR_BUFFER );
	tempRecvVar = (char *)vcp_malloc( strlen(RecvVar)+1 );
	if (( !tempCommand )|| ( !tempFileOrBuffer )|| ( !tempRecvVar ))
		return(OutOfMemory);
	strcpy( tempCommand, Command );
	strcpy( tempRecvVar, RecvVar );
	if ( TimeOutCommand != NULL ){
		tempTimeOutCommand = (char *)vcp_malloc( strlen(TimeOutCommand)+1 );
		if ( !tempTimeOutCommand )
			return(OutOfMemory);
		strcpy( tempTimeOutCommand, TimeOutCommand );
	}
	else
		tempTimeOutCommand = NULL;
	*tempFileOrBuffer = *FileOrBuffer;
	if ( FileOrBuffer->FileFlag ) {
		tempFileName = (char *)vcp_malloc( strlen(FileOrBuffer->BufferPtr)+1 );
		strcpy( tempFileName, FileOrBuffer->BufferPtr );
		tempFileOrBuffer->BufferPtr = tempFileName;
	}
	pid = ProcAlloc( Chan_ReceiveThread, DEFAULT_STACKSIZE, 
		7, LinkNumber, tempFileOrBuffer, count,
		tempCommand, tempRecvVar, TimeOut, tempTimeOutCommand  );
	ProcRun( pid );
	return(VCP_OK);
}
/*}}}*/

/*{{{  Chan_BlockedSend*/

/*
 *------------------------------------------------------------------------
 *
 *	Chan_BlockedSend :
 *		Send a message from a file or a specified buffer. The call
 *	blocks until the complete message is sent and the acknowledgement
 *	received.
 *	
 *	Parameters :
 *		LinkNumber :
 *			Link on which message is to be sent.
 *		FileOrBuffer :
 *			FileFlag = True implies send from a file. In this
 *			case, BufferPtr specifies the filename.
 *			If FileFlag = False, the data in BufferPtr is sent.
 *			Data in buffer is of the form (<HexByte> )*
 *		TimeOut :
 *			Number of microseconds to wait for message to be sent.
 *		TimeOutCommand :
 *			TCL Command to be executed on timeout
 *
 *	Returns nothing
 *
 *------------------------------------------------------------------------
 */
int Chan_BlockedSend( int LinkNumber, FILE_OR_BUFFER *FileOrBuffer , int TimeOut, char *TimeOutCommand)
{
	FILE	*fp;
	int	bufsize;
	unsigned char	*tempBuffer;
	int	MessageLength = 0;
	Boolean EOMFlag;

	if (( VCP_Initialized )&&( VCP_Running[LinkNumber] ))
		return(VCP_AlreadyRunning);
	if ( !FileOrBuffer->FileFlag ) {
		bufsize = strlen(FileOrBuffer->BufferPtr);
		MAKE_MESSAGE( tempBuffer, FileOrBuffer->BufferPtr, bufsize, MessageLength );
		fprintf(stderr, "Send %d bytes on link %d\n", MessageLength, LinkNumber );
		if ( TimeOut > 0 ){
			if (ChanOutTimeFail( OutChannel[LinkNumber], tempBuffer, MessageLength, TimeOut ) != 0 )
				QUEUE_TCL_COMMAND( TimeOutCommand );
		}
		else
			ChanOut( OutChannel[LinkNumber], tempBuffer, MessageLength );
	}
	else {
		EOMFlag = False;
		tempBuffer = ( unsigned char *)vcp_malloc( MAX_BUF_SIZE );
		if ( tempBuffer == NULL )
			return( OutOfMemory );
		fp = fopen( FileOrBuffer->BufferPtr, "r" );
		do {
			bufsize = fread( tempBuffer, sizeof(char), MAX_BUF_SIZE, fp );
			if ( bufsize <= 0 )
				break;
			if ( feof(fp))
				EOMFlag = True;
			if ( TimeOut > 0 ){
				if (ChanOutTimeFail( OutChannel[LinkNumber], tempBuffer, bufsize, TimeOut ) != 0 ){
					QUEUE_TCL_COMMAND( TimeOutCommand );
					break;
				}
			}
			else
				ChanOut( OutChannel[LinkNumber], tempBuffer, bufsize );
		} while ( EOMFlag != True );
		fclose(fp);
		vcp_free( tempBuffer );
	}
	return(VCP_OK);
}
/*}}}*/

/*{{{  Chan_SendThread*/

/*
 *------------------------------------------------------------------------
 *
 *	Chan_SendThread :
 *		Invokes Chan_BlockedSend to send a message from a file
 *	or buffer. When the send is complete, this thread executes the
 *	TCL command specified in the parameter Command, and then terminates.
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
static void Chan_SendThread( int FIXME, int LinkNumber, 
	FILE_OR_BUFFER *FileOrBuffer, char *Command , int TimeOut, char *TimeOutCommand)
{
	TCL_COMMAND	*TclCommand;

	Chan_BlockedSend( LinkNumber, FileOrBuffer, TimeOut, TimeOutCommand  );
	QUEUE_TCL_COMMAND( Command );
	if ( FileOrBuffer->FileFlag )
		vcp_free( FileOrBuffer->BufferPtr );
	vcp_free( FileOrBuffer );
	if ( TimeOutCommand != NULL )vcp_free( TimeOutCommand ); vcp_free( Command );
}
/*}}}*/

/*{{{  Chan_UnBlockedSend*/
/*
 *------------------------------------------------------------------------
 *
 *	Chan_UnBlockedSend :
 *		Saves all the parameters and creates a thread to send
 *	a message
 *
 *	Returns nothing.
 *
 *------------------------------------------------------------------------
 */
int Chan_UnBlockedSend( int LinkNumber, FILE_OR_BUFFER *FileOrBuffer, 
	char *Command , int TimeOut, char *TimeOutCommand )
{
	Process *pid;
	char *tempCommand;
	FILE_OR_BUFFER *tempFileOrBuffer;
	char *tempFileName;
	char *tempTimeOutCommand;

	if (( VCP_Initialized )&&( VCP_Running[LinkNumber] ))
		return(VCP_AlreadyRunning);
	tempCommand = (char *)vcp_malloc( strlen(Command)+1 );
	tempFileOrBuffer = MEM_ALLOC( FILE_OR_BUFFER );
	if (( !tempCommand )||( !tempFileOrBuffer ))
		return(OutOfMemory);
	if ( TimeOutCommand != NULL ){
		tempTimeOutCommand = (char *)vcp_malloc( strlen(TimeOutCommand)+1 );
		if ( !tempTimeOutCommand )
			return(OutOfMemory);
		strcpy( tempTimeOutCommand, TimeOutCommand );
	}
	else
		tempTimeOutCommand = NULL;
	strcpy( tempCommand, Command );
	*tempFileOrBuffer = *FileOrBuffer;
	if ( FileOrBuffer->FileFlag ) {
		tempFileName = (char *)vcp_malloc( strlen(FileOrBuffer->BufferPtr)+1 );
		strcpy( tempFileName, FileOrBuffer->BufferPtr );
		tempFileOrBuffer->BufferPtr = tempFileName;
	}
	pid = ProcAlloc( Chan_SendThread, DEFAULT_STACKSIZE, 
		6, LinkNumber, tempFileOrBuffer, tempCommand, TimeOut, tempTimeOutCommand );
	ProcRun( pid );
	return(VCP_OK);
}

/*}}}*/
