/*	Macro definitions used by VCP
*/

#ifndef	VCP_VCPMAC_H
#define VCP_VCPMAC_H

/*{{{  Access macros*/
/*	Return True if EOM packet, False otherwise
*/
#define PACKET_TYPE( count )	(( count >> EOM_BIT_NUMBER ) & 0x1)

/*	Return length of a packet
*/
#define	PACKET_LENGTH( packet ) ( packet[0] & COUNT_MASK )

/*	Return pointer to packet header
*/
#define PACKET_HEADER( packet ) ( &packet[1] )

/*	Return a 16 bit value from that pointer.
	NOTE : It is assumed that data representation is Little-endian
*/
#define GET_SHORT( ptr )	( (*(ptr)) + (*(ptr+1)) * 256 )

/*	Packet terminated by EOP ?
*/
#define IS_EOP_PACKET(packet)	( PACKET_TYPE(*(unsigned char *)packet) == 0 )

/*	Packet terminated by EOM ?
*/
#define IS_EOM_PACKET(packet)	( PACKET_TYPE(*(unsigned char *)packet) == 1 )

/*	Structures for a particular link: used by sender and receiver threads
*/
#define	myAckToBeSent 	 	AckToBeSent[LinkNumber]
#define	mySendRequestQ 	 	SendRequestQ[LinkNumber]
#define	myPendingSendRequests 	PendingSendRequests[LinkNumber]
#define	myTimerRequestQ 	TimerRequestQ[LinkNumber]
#define myPacketsReceived  	PacketsReceived[LinkNumber]
#define myVirtualChanInfo  	VirtualChanInfo[LinkNumber]

/*	Is ACK expected for this virtual channel?
*/
#define ACK_DUE( VChan )	(VChan->AckExpected.Flag == True)

/*	Is this an ACK packet ?
*/
#define	IsACKNOWLEDGEMENT( packet )					\
	( IS_EOP_PACKET( packet )&&( PACKET_LENGTH( packet ) < MAX_PACKET_DATA ))
/*}}}*/
	
/*{{{  Memory allocation macros*/
/*	Allocate memory for one element of specified type
*/
#define	MEM_ALLOC( type ) 	( type *)vcp_malloc( sizeof( type ))

/*	Allocate memory for one element of the specified type. 
	In case of error, error number is returned in Error,
	and the action specified in ReturnStatus is executed
*/
#define	MEM_ALLOCATE( var, type, Error, ReturnStatus ) {		\
	var = ( type *)vcp_malloc( sizeof(type));				\
	if ( var == NULL ){						\
		Error = VCP_ERROR( OutOfMemory );			\
		ReturnStatus;						\
	}								\
}
/*}}}*/

/*{{{  TRANSMIT_MESSAGE*/
/*	One packet from SendReq is sent over the specified channel.
	Format of message sent : <EOM><Count> <header> <packet>
	SendReq is updated to reflect the new buffer pointer and size,
	and AckExpected flag is set. Any error is returned in ErrorFlag.
*/
#ifdef TEST
#define TRANSMIT_MESSAGE( Chan, header, SendReq, ErrorFlag ){		\
	unsigned char *tempBuf;						\
	int templen = SendReq->BufferSize;				\
	int tempEOM = SendReq->EOMFlag;					\
	char *MsgBuf = SendReq->BufferPtr;				\
	int MsgSize;							\
	int size;							\
	MsgSize = (templen > MAX_PACKET_DATA)?MAX_PACKET_DATA:templen;	\
	SendReq->BufferSize -= MsgSize;					\
	SendReq->BufferPtr += MsgSize;					\
	size = header.HeaderSize+1+MsgSize;				\
	tempBuf = ( unsigned char *)vcp_malloc( size );			\
	if ( tempBuf == NULL ){						\
		ErrorFlag = VCP_ERROR( OutOfMemory );			\
	}								\
	else {								\
		tempEOM = (tempEOM && (MsgSize == templen))?1:0;	\
		tempEOM <<= EOM_BIT_NUMBER;				\
		tempEOM &= EOM_MASK;					\
		templen = MsgSize + header.HeaderSize;			\
		templen |= tempEOM;					\
		tempBuf[0] = (unsigned char )templen;			\
		memcpy( &tempBuf[1], header.HeaderPtr, header.HeaderSize );\
		memcpy( &tempBuf[1+header.HeaderSize], MsgBuf, MsgSize );\
		size = ( size > MIN_PACKET_SIZE)?size:MIN_PACKET_SIZE;	\
		ChanOut( Chan, tempBuf, MIN_PACKET_SIZE );		\
		size -= MIN_PACKET_SIZE;				\
		if ( size > 0 )						\
			ChanOut( Chan, (tempBuf+MIN_PACKET_SIZE), size );\
		vcp_free( tempBuf );					\
	}								\
}
#else
#define TRANSMIT_MESSAGE( Chan, header, SendReq, ErrorFlag ){		\
	unsigned char tempLen;						\
	int tempEOM = SendReq->EOMFlag;					\
	char *MsgBuf = SendReq->BufferPtr;				\
	int MsgSize = SendReq->BufferSize;				\
	int TimeOut = SendReq->TimeOut;					\
	int ret;							\
	MsgSize = (MsgSize > MAX_PACKET_DATA)?MAX_PACKET_DATA:MsgSize;	\
	SendReq->BufferSize -= MsgSize;					\
	SendReq->BufferPtr += MsgSize;					\
	tempEOM = (tempEOM && (SendReq->BufferSize == 0))?1:0;		\
	tempEOM = ( tempEOM << EOM_BIT_NUMBER ) & EOM_MASK;		\
	tempLen = MsgSize + header.HeaderSize;				\
	tempLen |= tempEOM;						\
	if ( TimeOut <= 0 ) {						\
		ChanOut( Chan, &tempLen, sizeof(char));			\
		ChanOut( Chan, header.HeaderPtr, header.HeaderSize );	\
		if ( MsgSize > 0 )					\
			ChanOut( Chan, MsgBuf, MsgSize );		\
	}								\
	else {								\
		ret=ChanOutTimeFail( Chan, &tempLen, sizeof(char), TimeOut);\
		if ( ret == 0 )						\
			ChanOutTimeFail( Chan, header.HeaderPtr, 	\
				header.HeaderSize, TimeOut );		\
		if ((ret ==0)&&( MsgSize > 0 ))				\
			ChanOutTimeFail( Chan, MsgBuf, MsgSize, TimeOut );\
	}								\
}
#endif
/*}}}*/
	
/*{{{  Copy message*/
/*	Copy a message received to user's buffer
*/
#define	CopyMessage	memcpy


/*	Is the user's buffer full
*/
#define	BUFFER_FULL( ReceiveReq ) ( ReceiveReq->MessageLength >= ReceiveReq->BufferSize )
/*}}}*/

/*{{{  SET_AckToBeSent*/
/*	Queue an ACK request, and wake up sender thread if required
*/
#define SET_AckToBeSent( LinkNumber, ReceiveReq ) {			\
	ACK_TO_BE_SENT *tempAck;					\
	tempAck = MEM_ALLOC( ACK_TO_BE_SENT );				\
	if ( tempAck == NULL )						\
		VCP_ERROR( OutOfMemory );				\
	else {								\
		ReceiveReq->VirtualChanInfo->AckNotYetSent = True;	\
		tempAck->ReceiveRequest = ReceiveReq;			\
		tempAck->TimeOut = ReceiveReq->TimeOut;			\
		EnQueue( AckToBeSent[LinkNumber], tempAck );		\
		if ( SenderWaitStruct[LinkNumber].SleepFlag == True ){	\
			SenderWaitStruct[LinkNumber].SleepFlag = False;	\
			ProcAwaken( SenderWaitStruct[LinkNumber].ProcessDescriptor );\
		}							\
	}								\
}
/*}}}*/

/*{{{  SEND_ACK*/
#ifndef TEST
/*	Send an acknowledgement to the specified virtual channel
*/
#define	SEND_ACK( Chan, header, TimeOut, ErrorFlag ){			\
	char *tempBuf;							\
	tempBuf = (char *)vcp_malloc(header.HeaderSize+1);		\
	tempBuf[0] = header.HeaderSize;					\
	memcpy( &tempBuf[1], header.HeaderPtr, header.HeaderSize );	\
	if ( TimeOut <= 0 )						\
		ChanOut( Chan, tempBuf, header.HeaderSize+1 );		\
	else								\
		ChanOutTimeFail( Chan, tempBuf, header.HeaderSize+1, TimeOut );\
	vcp_free( tempBuf );						\
}
#else
#define	SEND_ACK( Chan, header, TimeOut, ErrorFlag ){			\
	char tempBuf[MIN_PACKET_SIZE];					\
	tempBuf[0] = header.HeaderSize;					\
	memcpy( &tempBuf[1], header.HeaderPtr, header.HeaderSize );	\
	if ( TimeOut <= 0 )						\
		ChanOut( Chan, tempBuf, MIN_PACKET_SIZE );		\
	else								\
		ChanOutTimeFail( Chan, tempBuf, MIN_PACKET_SIZE, TimeOut );\
}

#endif
/*}}}*/

/*{{{  GET_VIRTUAL_CHAN_INFO*/
/*	If there is a virtual channel corresponding to the header in
	the given packet, return a pointer to the virtual channel
	info in VChan.
*/
#define	GET_VIRTUAL_CHAN_INFO( VChanInfo, packet, VChan ){		\
	VIRTUAL_CHAN tempHeader;					\
	tempHeader.HeaderSize = 1;					\
	tempHeader.HeaderValue = *(char *)packet;			\
	GET_VCHAN_FROM_TABLE( VChanInfo, tempHeader, VChan );		\
	if ( VChan == NULL ) {						\
		tempHeader.HeaderSize = 2;				\
		tempHeader.HeaderValue = GET_SHORT(packet);		\
		GET_VCHAN_FROM_TABLE( VChanInfo, tempHeader, VChan );	\
	}								\
}
/*}}}*/

/*{{{  GET_VCHAN_FROM_TABLE*/
/*	If there is a virtual channel with the specified header in the
	table, return it in element, else return NULL
*/
#define GET_VCHAN_FROM_TABLE( table, header, element ){			\
	element = (VIRTUAL_CHAN_INFO *)table[header.HeaderValue % MAX_TABLE_SIZE];		\
	while (((void *) element != NULL )&&				\
	 (( element->ReceiveHeader.HeaderValue != header.HeaderValue )||\
	 ( element->ReceiveHeader.HeaderSize != header.HeaderSize )))	\
		element = QNext(element);				\
}
/*}}}*/

/*{{{  GET_PACKET_FROM_TABLE*/
/*	If a packet is available for the virtual channel specified by
	header, return the packet in element
*/
#define GET_PACKET_FROM_TABLE( table, header, element ){		\
	element = (PACKETS_RECEIVED *)table[header.HeaderValue % MAX_TABLE_SIZE];		\
	if ( header.HeaderSize == 1 ) {					\
		while (((void *) element != NULL )&&			\
		 ( *((char *)(element->PacketBuffer+1)) != header.HeaderValue ))\
			element = QNext(element);			\
	}								\
	else {								\
		while (((void *) element != NULL )&&			\
		 ( GET_SHORT(element->PacketBuffer+1) != header.HeaderValue ))\
			element = QNext(element);			\
	}								\
	if ((void *) element != NULL )						\
		RemoveTable( table, header.HeaderValue, element );	\
}
/*}}}*/

/*{{{  ADD / DELETE VCHAN*/
/*	Add virtual channel information for the specified pair of
	headers, if it is not present already
*/
#define	ADD_VCHAN( VChanInfo, SendHeader, ReceiveHeader, tempVChan ){	\
	GET_VCHAN_FROM_TABLE( VChanInfo, ReceiveHeader, tempVChan );	\
	if ( tempVChan == NULL ) {					\
		tempVChan = MEM_ALLOC( VIRTUAL_CHAN_INFO );		\
		if ( tempVChan == NULL )				\
			VCP_ERROR( OutOfMemory );			\
		else {							\
			tempVChan->IgnoreFlag = False;			\
			tempVChan->SendHeader = *SendHeader;		\
			tempVChan->SendHeader.HeaderPtr =(char *)	\
				vcp_malloc(SendHeader->HeaderSize);	\
			memcpy( tempVChan->SendHeader.HeaderPtr, 	\
				SendHeader->HeaderPtr, SendHeader->HeaderSize );\
			tempVChan->ReceiveHeader = ReceiveHeader;	\
			InitQ( tempVChan->PendingSendRequests );	\
			InitQ( tempVChan->ReceiveRequest );		\
			tempVChan->AckExpected.Flag = False;		\
			tempVChan->AckNotYetSent = False;		\
			AddTable( VChanInfo, tempVChan->ReceiveHeader.HeaderValue, tempVChan );	\
		}							\
	}								\
}

/*	Delete virtual channel information, if the channel is not 
	currently in use.
*/
#define	DELETE_VCHAN( VChanInfo, VChan ) {				\
	if ( VCHAN_NOT_BEING_USED( VChan )){				\
		vcp_free( VChan->SendHeader.HeaderPtr );		\
		DeleteTable( VChanInfo, 				\
			VChan->ReceiveHeader.HeaderValue, VChan );	\
	}								\
}
/*}}}*/

/*{{{  WAKEUP_RECEIVER*/
/*	Wake up user thread waiting for a receive, if either the user buffer
	is full or End Of Message has been received
*/
#define	WAKEUP_RECEIVER_IF_REQUIRED( VChanInfo, ReceiveReq ) {		\
	if (((void *) ReceiveReq != NULL )&&				\
		( ReceiveReq->EOMFlag || BUFFER_FULL( ReceiveReq ))) {	\
		QRemove( ReceiveReq->VirtualChanInfo->ReceiveRequest, 	\
			ReceiveReq );					\
		DELETE_VCHAN( VChanInfo, ReceiveReq->VirtualChanInfo );	\
		if ( ReceiveReq->TimeOut > 0 ){				\
			REMOVE_TIMER_REQUEST( ReceiveReq->TimerRequest, \
				TimerChannel );				\
		}							\
		ProcAwaken( ReceiveReq->ReceiverProcess );		\
		/*vcp_free( ReceiveReq );--CHECK : freed by receiver*/	\
	}								\
}
/*}}}*/

/*{{{  REMOVE_TIMER_REQUEST*/
/*	Delete a timer request and wakeup the timer thread
*/
#define REMOVE_TIMER_REQUEST( TimerReq, TimerChannel ){			\
	int temp;							\
	QDelete( myTimerRequestQ, TimerReq );				\
	SemWait( &TimerSemaphore[LinkNumber] );				\
	ChanOutInt( TimerChannel, temp );				\
	SemSignal( &TimerSemaphore[LinkNumber] );			\
}
/*}}}*/

/*{{{  SET_RECEIVE_TIMER*/
/*	Set a timer for a receive request. Wakeup the timer thread
*/
#define SET_RECEIVE_TIMER( ReceiveReq, TimeOut ){			\
	TIMER_REQUEST *tempTimerReq;					\
	int ErrorStatus;						\
	if ( TimeOut > 0 ){						\
		MEM_ALLOCATE( tempTimerReq, TIMER_REQUEST, 		\
			ErrorStatus, return(VCP_ERROR(OutOfMemory) ));	\
		tempTimerReq->DeadLine = ProcTime()+TimeOut;		\
		tempTimerReq->IsSendReq = False;			\
		tempTimerReq->request.ReceiveRequest = ReceiveReq;	\
		ReceiveReq->TimerRequest = tempTimerReq;		\
		ADD_TIMER_Q( myTimerRequestQ, tempTimerReq );		\
		SemWait( &(TimerSemaphore[LinkNumber]) );		\
		ChanOutInt( Timer_ReceiverChannel[LinkNumber],TimeOut );\
		SemSignal( &(TimerSemaphore[LinkNumber]) );		\
	}								\
}
/*}}}*/

/*{{{  POST_REQUEST*/
/*	Queue a receive request on the specified virtual channel
*/
#define POST_REQUEST( buffer,  bufsize, VChan, ReceiveReq, timeOut ){	\
	ReceiveReq = MEM_ALLOC(RECEIVE_REQUEST);			\
	if ((void *) ReceiveReq != NULL ){				\
		ReceiveReq->VirtualChanInfo = VChan;			\
		ReceiveReq->BufferSize = bufsize;			\
		ReceiveReq->BufferPtr  = buffer;			\
		ReceiveReq->MessageLength = 0;				\
		ReceiveReq->EOMFlag  = False;				\
		ReceiveReq->Error = VCP_OK;				\
		ReceiveReq->TimeOut = timeOut;				\
		ReceiveReq->SleepFlag = False;				\
		EnQueue( VChan->ReceiveRequest, ReceiveReq );		\
		SET_RECEIVE_TIMER( ReceiveReq, timeOut );		\
	}								\
	else return( VCP_ERROR(OutOfMemory));				\
}
/*}}}*/

/*{{{  ADD_TO_MESSAGE_LIST*/
/*	Queue a message when ReceiveAll is set
*/
#define	ADD_TO_MESSAGE_LIST( MessageList, Packet )			\
	EnQueue( MessageList, Packet )
/*}}}*/

/*{{{  VCHAN_NOT_BEING_USED*/
/*	Is VChan in use ?
*/
#define VCHAN_NOT_BEING_USED( VChan )					\
	(( QEmpty( VChan->PendingSendRequests ))&&			\
	 ( QEmpty( VChan->ReceiveRequest ))&&				\
	 ( VChan->AckExpected.Flag == False )&&				\
	 ( VChan->AckNotYetSent == False ))
/*}}}*/

/*{{{  ADD_TIMER_Q*/
/*	Add a timer request to the timer request Q. The requests are
	ordered by theie deadlines
*/
#define ADD_TIMER_Q( TimerQ, TimerReq ) {				\
	TIMER_REQUEST *prev, *next;					\
	next = TimerQ; prev = NULL;					\
	if ( QEmpty(TimerQ))						\
		 InsertQ( TimerQ, TimerReq, prev )			\
	else {								\
		while ( next != NULL ){					\
			if ( next->DeadLine > TimerReq->DeadLine ){	\
				InsertQ( TimerQ, TimerReq, prev );	\
				break;					\
			}						\
			prev = next; next = QNext(next); 		\
		}							\
		if ( next == NULL ) InsertQ( TimerQ, TimerReq, prev );	\
	}								\
}
/*}}}*/
		
/*{{{  ABORT SEND/RECV REQUEST*/
/*	Abort a send request, wakeup the thread which is sleeping on the
	request
*/
#define ABORT_SEND_REQUEST( SendReq ){					\
	VIRTUAL_CHAN_INFO *tempVChan;					\
	tempVChan = SendReq->VirtualChanInfo;				\
	tempVChan->AckExpected.Flag = False;				\
	DELETE_VCHAN( myVirtualChanInfo, tempVChan );			\
	SendReq->Error = VCP_TimeOut;					\
	ProcAwaken( SendReq->SenderProcess );				\
}

/*	Abort a receive request, wakeup the thread which is sleeping on
	the request
*/
#define ABORT_RECV_REQUEST( RecvReq ){					\
	VIRTUAL_CHAN_INFO *tempVChan;					\
	tempVChan = RecvReq->VirtualChanInfo;				\
	QRemove( tempVChan->ReceiveRequest, RecvReq );			\
	DELETE_VCHAN( myVirtualChanInfo, tempVChan );			\
	RecvReq->Error = VCP_TimeOut;					\
	ProcAwaken( RecvReq->ReceiverProcess );				\
}
/*}}}*/
#endif
